# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import jax
from vllm.logger import init_logger
from vllm.utils import (get_distributed_init_method, get_ip, get_open_port,
                        run_method)
from vllm.v1.executor.abstract import Executor
from vllm.worker.worker_base import WorkerWrapperBase

logger = init_logger(__name__)


class DisaggExecutor(Executor):

    def _init_executor(self) -> None:
        """Initialize the worker and load the model.
        """
        self.driver_worker = WorkerWrapperBase(vllm_config=self.vllm_config,
                                               rpc_rank=0)
        slice_config = getattr(self.vllm_config.device_config, "slice")
        idx = slice_config[0]
        sizes = slice_config[1]

        start = sum(sizes[0:idx])
        end = start + sizes[idx]

        devices = jax.devices()[start:end]
        setattr(self.vllm_config.device_config, "slice", (idx + 1, sizes))
        logger.info(
            f"Creating DisaggExecutor with {devices}, index: {start} -> {end}")

        distributed_init_method = get_distributed_init_method(
            get_ip(), get_open_port())
        local_rank = 0
        rank = 0
        is_driver_worker = True
        kwargs = dict(
            vllm_config=self.vllm_config,
            local_rank=local_rank,
            rank=rank,
            distributed_init_method=distributed_init_method,
            is_driver_worker=is_driver_worker,
            devices=devices,
        )
        self.collective_rpc("init_worker", args=([kwargs], ))
        self.collective_rpc("init_device")
        self.collective_rpc("load_model")

    def collective_rpc(self,
                       method: Union[str, Callable],
                       timeout: Optional[float] = None,
                       args: Tuple = (),
                       kwargs: Optional[Dict] = None) -> List[Any]:
        if kwargs is None:
            kwargs = {}
        answer = run_method(self.driver_worker, method, args, kwargs)
        return [answer]

    def check_health(self) -> None:
        # DisaggExecutor will always be healthy as long as
        # it's running.
        return



# SPDX-License-Identifier: Apache-2.0

import jax
import jax.numpy as jnp
import numpy as np
from vllm.logger import init_logger
from vllm.utils import cdiv

logger = init_logger(__name__)


# TODO(xiang): fix device allocation
class BlockTable:

    def __init__(
        self,
        max_num_reqs: int,
        max_num_blocks_per_req: int,
        max_num_batched_tokens: int,
        pin_memory: bool,
    ):
        self.max_num_reqs = max_num_reqs
        self.max_num_blocks_per_req = max_num_blocks_per_req
        self.max_num_batched_tokens = max_num_batched_tokens
        self.pin_memory = pin_memory

        self.block_table = jnp.zeros(
            (max_num_reqs, max_num_blocks_per_req),
            dtype=jnp.int32,
        )
        self.block_table_cpu = np.zeros(
            (max_num_reqs, max_num_blocks_per_req),
            dtype=jnp.int32,
        )
        self.num_blocks_per_row = np.zeros(max_num_reqs, dtype=np.int32)

    def append_row(
        self,
        block_ids: list[int],
        row_idx: int,
    ) -> None:
        if not block_ids:
            return
        num_blocks = len(block_ids)
        start = self.num_blocks_per_row[row_idx]
        self.num_blocks_per_row[row_idx] += num_blocks
        self.block_table_cpu[row_idx, start:start + num_blocks] = block_ids

    def add_row(self, block_ids: list[int], row_idx: int) -> None:
        self.num_blocks_per_row[row_idx] = 0
        self.append_row(block_ids, row_idx)

    def move_row(self, src: int, tgt: int) -> None:
        num_blocks = self.num_blocks_per_row[src]
        self.block_table_cpu[tgt, :num_blocks] = self.block_table_cpu[
            src, :num_blocks]
        self.num_blocks_per_row[tgt] = num_blocks

    def swap_row(self, src: int, tgt: int) -> None:
        num_blocks_src = self.num_blocks_per_row[src]
        num_blocks_tgt = self.num_blocks_per_row[tgt]
        self.num_blocks_per_row[src] = num_blocks_tgt
        self.num_blocks_per_row[tgt] = num_blocks_src

        self.block_table_cpu[[src, tgt]] = self.block_table_cpu[[tgt, src]]

    def commit(self, num_reqs: int) -> None:
        self.block_table[:num_reqs].copy_(self.block_table_cpu[:num_reqs],
                                          non_blocking=True)

    def clear(self) -> None:
        self.block_table.fill_(0)
        self.block_table_cpu.fill_(0)

    def get_device_tensor(self) -> jax.Array:
        """Ruturns the device tensor of the block table."""
        return self.block_table

    def get_cpu_tensor(self) -> jax.Array:
        """Returns the CPU tensor of the block table."""
        return self.block_table_cpu


class MultiGroupBlockTable:
    """The BlockTables for each KV cache group."""

    def __init__(self, max_num_reqs: int, max_model_len: int,
                 max_num_batched_tokens: int, pin_memory: bool,
                 block_sizes: list[int]) -> None:
        self.block_tables = [
            BlockTable(max_num_reqs, cdiv(max_model_len, block_size),
                       max_num_batched_tokens, pin_memory)
            for block_size in block_sizes
        ]

    def append_row(self, block_ids: list[list[int]], row_idx: int) -> None:
        for i, block_table in enumerate(self.block_tables):
            block_table.append_row(block_ids[i], row_idx)

    def add_row(self, block_ids: list[list[int]], row_idx: int) -> None:
        for i, block_table in enumerate(self.block_tables):
            block_table.add_row(block_ids[i], row_idx)

    def move_row(self, src: int, tgt: int) -> None:
        for block_table in self.block_tables:
            block_table.move_row(src, tgt)

    def swap_row(self, src: int, tgt: int) -> None:
        for block_table in self.block_tables:
            block_table.swap_row(src, tgt)

    def commit(self, num_reqs: int) -> None:
        for block_table in self.block_tables:
            block_table.commit(num_reqs)

    def clear(self) -> None:
        for block_table in self.block_tables:
            block_table.clear()

    def __getitem__(self, idx: int) -> "BlockTable":
        """Returns the BlockTable for the i-th KV cache group."""
        return self.block_tables[idx]



import unittest

from tpu_commons.core.disagg_utils import _parse_slices


class DisaggUtilsTest(unittest.TestCase):

    def test_parse_slices_valid_cases(self):
        """Tests valid slice strings."""
        # Test with a single slice
        self.assertEqual(_parse_slices("2x2"), (4, ))
        self.assertEqual(_parse_slices("2"), (2, ))

        # Test with multiple slices
        self.assertEqual(_parse_slices("2x2,2x1,3,2x4"), (4, 2, 3, 8))

        # Test with various dimensions
        self.assertEqual(_parse_slices("1x1,10x10,5x3"), (1, 100, 15))

        # Test with an empty string
        self.assertEqual(_parse_slices(""), ())

    def test_parse_slices_with_whitespace(self):
        """Tests valid slice strings with extra whitespace."""
        self.assertEqual(_parse_slices(" 2x2 "), (4, ))
        self.assertEqual(_parse_slices(" 2x2 , 2x1 , 2x4 "), (4, 2, 8))
        # The current implementation allows spaces inside the slice definition
        # because int() can handle them.
        self.assertEqual(_parse_slices("2 x 2"), (4, ))
        self.assertEqual(_parse_slices(" 10 x 10 "), (100, ))

    def test_parse_slices_invalid_cases(self):
        """Tests malformed slice strings that should raise ValueError."""
        invalid_strings = [
            "2*2",  # wrong separator
            "2x",  # incomplete
            "axb",  # not integers
            "2x2x2",  # too many dimensions
            "2x2,3*3",  # partially malformed
            ",2x2",  # leading comma
            "2x2,",  # trailing comma
            "2x2,,2x1",  # empty slice in middle
        ]
        for invalid_str in invalid_strings:
            with self.subTest(invalid_str=invalid_str):
                with self.assertRaises(ValueError):
                    _parse_slices(invalid_str)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import functools

import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized
from jax._src import test_util as jtu

from tpu_commons.kernels.quantized_matmul import (kernel, tuned_block_sizes,
                                                  util)

quantized_matmul_kernel = kernel.quantized_matmul_kernel
quantize_tensor = util.quantize_tensor
get_tuned_block_sizes = tuned_block_sizes.get_tuned_block_sizes

jax.config.parse_flags_with_absl()


@functools.partial(jax.jit, static_argnames=["quantize_activation"])
def reference_quantized_matmul(x, w_q, w_scale, quantize_activation=True):
    if quantize_activation:
        x_q, x_scale = quantize_tensor(x)
        out = jax.lax.dot_general(
            x_q,
            w_q,
            dimension_numbers=(((1, ), (1, )), ((), ())),
            preferred_element_type=jnp.int32,
        ).astype(jnp.float32)
        out *= x_scale
    else:
        out = jax.lax.dot_general(
            x,
            w_q,
            dimension_numbers=(((1, ), (1, )), ((), ())),
            preferred_element_type=jnp.float32,
        )
    out *= jnp.expand_dims(w_scale, 0)
    return out.astype(x.dtype)


@jtu.with_config(jax_numpy_dtype_promotion="standard")
class QuantizedMatmulKernelTest(jtu.JaxTestCase):

    def setUp(self):
        super().setUp()
        if not jtu.is_device_tpu_at_least(6):
            self.skipTest("Expect TPUv6+")

    def _test_quantized_matmul(
        self,
        dtype,
        bs,
        n_input_features,
        n_output_features,
        quantize_activation,
        batch_block_size=None,
        out_block_size=None,
        in_block_size=None,
        atol=2e-1,
        rtol=1e-1,
    ):

        prng_key = jax.random.key(1234)
        k0, k1 = jax.random.split(prng_key, 2)
        x = jax.random.uniform(k0, (bs, n_input_features),
                               dtype=dtype,
                               minval=0,
                               maxval=1)
        w = jax.random.uniform(
            k1,
            (n_output_features, n_input_features),
            dtype=dtype,
            minval=-1,
            maxval=1,
        )
        w_q, w_scale = quantize_tensor(w)
        w_scale = jnp.squeeze(w_scale)
        assert w_scale.shape == (n_output_features, )

        output = quantized_matmul_kernel(
            x,
            w_q,
            w_scale,
            quantize_activation=quantize_activation,
            batch_block_size=batch_block_size,
            out_block_size=out_block_size,
            in_block_size=in_block_size,
        )
        expected = reference_quantized_matmul(
            x, w_q, w_scale, quantize_activation=quantize_activation)

        self.assertAllClose(output,
                            expected,
                            rtol=rtol,
                            atol=atol,
                            check_dtypes=True)

    @parameterized.product(
        dtype=[jnp.bfloat16, jnp.float32],
        bs=[128, 256, 512],
        n_input_features=[128, 256, 512],
        n_output_features=[128, 256, 512],
        quantize_activation=[True],
    )
    def test_quantized_matmul_various_input_shapes(self, dtype, bs,
                                                   n_input_features,
                                                   n_output_features,
                                                   quantize_activation):
        self._test_quantized_matmul(
            dtype,
            bs,
            n_input_features,
            n_output_features,
            quantize_activation=quantize_activation,
            batch_block_size=128,
            out_block_size=128,
            in_block_size=128,
        )

    @parameterized.product(
        dtype=[jnp.bfloat16, jnp.float32],
        bs=[64, 192],
        n_input_features=[64, 192],
        n_output_features=[64, 192],
        quantize_activation=[True],
    )
    def test_quantized_matmul_unaligned_input_shapes(self, dtype, bs,
                                                     n_input_features,
                                                     n_output_features,
                                                     quantize_activation):
        self._test_quantized_matmul(
            dtype,
            bs,
            n_input_features,
            n_output_features,
            quantize_activation=quantize_activation,
            batch_block_size=128,
            out_block_size=128,
            in_block_size=128,
        )

    @parameterized.parameters(
        (jnp.bfloat16, 128, 1280, 8192, True),
        (jnp.bfloat16, 128, 28672, 4096, True),
        (jnp.bfloat16, 128, 4096, 14336, True),
        (jnp.bfloat16, 128, 4096, 4096, True),
        (jnp.bfloat16, 128, 6144, 4096, True),
        (jnp.bfloat16, 128, 7168, 8192, True),
        (jnp.bfloat16, 128, 8192, 1024, True),
        (jnp.bfloat16, 128, 8192, 3584, True),
    )
    def test_quantized_matmul_use_tuned_block_sizes(self, dtype, bs,
                                                    n_input_features,
                                                    n_output_features,
                                                    quantize_activation):
        self._test_quantized_matmul(
            dtype,
            bs,
            n_input_features,
            n_output_features,
            quantize_activation=quantize_activation,
            batch_block_size=None,
            out_block_size=None,
            in_block_size=None,
        )


if __name__ == "__main__":
    absltest.main(testLoader=jtu.JaxTestLoader())



from typing import List

import jax
import jax.numpy as jnp
from jax.sharding import Mesh, NamedSharding, PartitionSpec

import tpu_commons.kernels.ragged_paged_attention.v3.kernel as rpa
from tpu_commons.logger import init_logger

logger = init_logger(__name__)


def create_kv_caches(
    num_blocks: int,
    block_size: int,
    num_kv_heads: int,
    head_size: int,
    mesh: Mesh,
    layer_names: List[str],
    cache_dtype: jnp.dtype = jnp.bfloat16,
) -> List[jax.Array]:
    """
    Creates the KV caches, a list of arrays, each array is for one attention layer.

    The shape of the KV cache per layer is:
    (num_blocks, block_size, cdiv(num_kv_heads * 2, packing), packing, head_size).
    packing =  (32 // dtype bits)

    Args:
        num_blocks: The number of blocks in the KV cache.
        block_size: The size of each block in the KV cache.
        num_kv_heads: The number of KV heads in the KV cache.
        head_size: The size of each head in the KV cache.
        mesh: The mesh to shard the KV caches across.
        layer_names: The names of the decoder layers in the model.

    Returns:
        A list of KV caches, one per each decoder layer in the model.

    """
    # TODO(xiang): fix this together with get_kv_cache_spec
    # cache_dtype = kv_cache_spec.dtype

    cache_shape = rpa.get_kv_cache_shape(num_blocks, block_size, num_kv_heads,
                                         head_size, cache_dtype)

    sharding = NamedSharding(mesh, PartitionSpec(None, None, "model"))

    def _allocate() -> jax.Array:
        return jnp.empty(
            shape=cache_shape,
            dtype=cache_dtype,
        )

    sharded_allocate = jax.jit(_allocate, out_shardings=sharding)
    kv_caches = []
    for _ in layer_names:
        kv_caches.append(sharded_allocate())
    return kv_caches



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
import functools
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import torch
from jax.sharding import Mesh
from torchax.interop import jax_view, torch_view
from vllm.attention.backends.abstract import (AttentionBackend, AttentionImpl,
                                              AttentionLayer, AttentionType)

from tpu_commons.logger import init_logger
from tpu_commons.models.jax.attention import attention
from tpu_commons.models.jax.attention_metadata import AttentionMetadata
# Register custom op dispatcher.
from tpu_commons.models.vllm.vllm_model_wrapper_context import \
    get_vllm_model_wrapper_context

logger = init_logger(__name__)

TPU_STR_DTYPE_TO_JAX_DTYPE = {
    "bfloat16": jnp.bfloat16,
    "fp8": jnp.float8_e4m3fn,
    "fp8_e4m3": jnp.float8_e4m3,
    "fp8_e5m2": jnp.float8_e5m2,
    "int8": jnp.int8,
}


class PallasAttentionBackend(AttentionBackend):

    @staticmethod
    def get_name() -> str:
        return "PALLAS_VLLM_V1"

    @staticmethod
    def get_impl_cls() -> type["PallasAttentionBackendImpl"]:
        return PallasAttentionBackendImpl


class PallasAttentionBackendImpl(AttentionImpl):

    def __init__(
        self,
        num_heads: int,
        head_size: int,
        scale: float,
        num_kv_heads: int,
        alibi_slopes: Optional[list[float]],
        sliding_window: Optional[int],
        kv_cache_dtype: str,
        logits_soft_cap: Optional[float] = None,
        attn_type: str = AttentionType.DECODER,
        kv_sharing_target_layer_name: Optional[int] = None,
        use_irope: bool = False,
    ) -> None:
        if use_irope:
            logger.warning_once(
                "Using irope in Pallas is not supported yet, it will fall back "
                "to global attention for long context.")
        self.num_heads = num_heads
        self.head_size = head_size
        self.scale = float(scale)
        self.num_kv_heads = num_kv_heads
        self.sliding_window = sliding_window
        self.logits_soft_cap = logits_soft_cap
        self.kv_sharing_target_layer_name = kv_sharing_target_layer_name

        self.num_queries_per_kv = self.num_heads // self.num_kv_heads
        if alibi_slopes is not None:
            raise NotImplementedError("Alibi slopes is not supported.")
        self.kv_cache_quantized_dtype = None
        if kv_cache_dtype != "auto":
            self.kv_cache_quantized_dtype = TPU_STR_DTYPE_TO_JAX_DTYPE.get(
                kv_cache_dtype.lower().strip())

        if attn_type != AttentionType.DECODER:
            raise NotImplementedError("Encoder self-attention and "
                                      "encoder/decoder cross-attention "
                                      "are not implemented for "
                                      "PallasAttentionBackendImpl")

    def quantize_kv(self, layer: AttentionLayer, key: jax.Array,
                    value: jax.Array):
        dtype_info = jnp.finfo(self.kv_cache_quantized_dtype)
        minval, maxval = float(dtype_info.min), float(dtype_info.max)
        # TODO(kyuyeunk): Add support for jnp.Tensor (layer._k_scale) scale
        key = key.astype(jnp.float32) / layer._k_scale_float
        key = jnp.clip(key, minval, maxval)
        key = key.astype(self.kv_cache_quantized_dtype)
        value = value.astype(jnp.float32) / layer._v_scale_float
        value = jnp.clip(value, minval, maxval)
        value = value.astype(self.kv_cache_quantized_dtype)

        return key, value

    def forward(
        self,
        layer: AttentionLayer,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        kv_cache: torch.Tensor,
        attn_metadata: AttentionMetadata,
        output: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if output_scale is not None:
            raise NotImplementedError(
                "fused output quantization is not yet supported for "
                "PallasAttentionBackendImpl")

        if kv_cache.numel():
            raise RuntimeError(
                "KV cache from vLLM Attention layer should be empty but has "
                "the size of %s.", kv_cache.numel())

        del kv_cache  # Use kv_cache from vllm wrapper context values instead.

        vllm_model_wrapper_context = get_vllm_model_wrapper_context()
        kv_cache_index = vllm_model_wrapper_context.layer_name_to_kvcache_index[
            layer.layer_name]
        kv_cache = vllm_model_wrapper_context.kv_caches[kv_cache_index]

        mesh = vllm_model_wrapper_context.mesh

        query, key, value = jax_view(query), jax_view(key), jax_view(value)
        q_scale = k_scale = v_scale = None
        if self.kv_cache_quantized_dtype:
            key, value = self.quantize_kv(layer, key, value)
            # TODO(kyuyeunk): Add query quantization support for w8a8 attention
            # q_scale = layer._q_scale
            k_scale = layer._k_scale_float
            v_scale = layer._v_scale_float

        new_kv_cache, outputs = _jax_attn_func(kv_cache, query, key, value,
                                               attn_metadata, mesh, self.scale,
                                               self.head_size, self.num_heads,
                                               self.num_kv_heads, q_scale,
                                               k_scale, v_scale)
        vllm_model_wrapper_context.kv_caches[kv_cache_index] = new_kv_cache

        return torch_view(outputs)


@functools.partial(
    jax.jit,
    static_argnums=(
        5, 6, 7, 8, 9, 10, 11, 12
    ),  # mesh, scale, head_size, num_heads, num_kv_heads, q_scale, k_scale, v_scale
    donate_argnums=(0, ),  # donate kv_cache
)
def _jax_attn_func(
    kv_cache: jax.Array,
    q: jax.Array,
    k: jax.Array,
    v: jax.Array,
    attention_metadata: AttentionMetadata,
    mesh: Mesh,
    scale: float,
    head_size: int,
    num_heads: int,
    num_kv_heads: int,
    q_scale: Optional[float] = None,
    k_scale: Optional[float] = None,
    v_scale: Optional[float] = None,
) -> Tuple[jax.Array, jax.Array]:
    del scale  # Unused for now, as the attention function applies a default scale.

    # Get shapes from vllm
    q_len, q_compute_dim = q.shape
    k_len, k_compute_dim = k.shape
    assert k.shape == v.shape
    assert q_compute_dim == head_size * num_heads
    assert k_compute_dim == head_size * num_kv_heads

    # Convert the shapes from vLLM's convetion to what the attention function expects
    # bs, num_heads, q_len, head_size
    q = q.reshape(q_len, num_heads, head_size)
    # bs, num_kv_heads, k_len, head_size
    k = k.reshape(k_len, num_kv_heads, head_size)
    v = v.reshape(k_len, num_kv_heads, head_size)

    new_kv_cache, outputs = attention(
        kv_cache,
        q,
        k,
        v,
        attention_metadata,
        mesh,
        q_scale=q_scale,
        k_scale=k_scale,
        v_scale=v_scale,
    )

    # Convert the shape back to vLLM's convention
    assert outputs.shape[0] == q_len
    assert outputs.shape[1] == num_heads
    assert outputs.shape[2] == head_size
    outputs = outputs.reshape(q_len, q_compute_dim)

    return new_kv_cache, outputs



import importlib
import unittest
from unittest.mock import patch


class TestPathwaysInit(unittest.TestCase):

    @patch.dict("os.environ", {"JAX_PLATFORMS": "proxy,cpu"})
    def test_VLLM_TPU_USING_PATHWAYS_enabled(self):
        """Test when JAX_PLATFORMS contains 'proxy'."""

        # Import vllm.envs to test the VLLM_TPU_USING_PATHWAYS logic
        import vllm.envs as envs

        # Reload the module to ensure fresh import
        importlib.reload(envs)

        # Check that VLLM_TPU_USING_PATHWAYS is True when JAX_PLATFORMS contains "proxy"
        self.assertTrue(envs.VLLM_TPU_USING_PATHWAYS)

    @patch.dict("os.environ", {"JAX_PLATFORMS": "cpu"})
    def test_VLLM_TPU_USING_PATHWAYS_not_enabled(self):
        """Test when JAX_PLATFORMS does not contain 'proxy'."""

        # Import vllm.envs to test the VLLM_TPU_USING_PATHWAYS logic
        import vllm.envs as envs

        # Reload the module to ensure fresh import
        importlib.reload(envs)

        # Check that VLLM_TPU_USING_PATHWAYS is False when JAX_PLATFORMS doesn't contain "proxy"
        self.assertFalse(envs.VLLM_TPU_USING_PATHWAYS)

    @patch.dict("os.environ", {"JAX_PLATFORMS": "PROXY,CPU"})
    def test_VLLM_TPU_USING_PATHWAYS_case_insensitive(self):
        """Test that JAX_PLATFORMS check is case insensitive."""

        # Import vllm.envs to test the VLLM_TPU_USING_PATHWAYS logic
        import vllm.envs as envs

        # Reload the module to ensure fresh import
        importlib.reload(envs)

        # Check that VLLM_TPU_USING_PATHWAYS is True even with uppercase "PROXY"
        self.assertTrue(envs.VLLM_TPU_USING_PATHWAYS)


if __name__ == "__main__":
    unittest.main()



import random

import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized
from jax._src import test_util as jtu

from tpu_commons.kernels.ragged_paged_attention.v2.kernel import (
    dynamic_validate_inputs, ragged_paged_attention,
    ref_ragged_paged_attention)

jax.config.parse_flags_with_absl()


def ceil_div(x, a):
    assert a != 0
    return (x + a - 1) // a


@jtu.with_config(jax_numpy_dtype_promotion="standard")
class PagedAttentionKernelTest(jtu.JaxTestCase):

    def _test_ragged_paged_attention(
        self,
        seq_lens,  # List[(q_len, kv_len)]
        num_heads,  # [num_q_heads, num_kv_heads]
        head_dim,
        page_size,
        dtype,
        num_pages,
        *,
        num_kv_pages_per_block=8,
        num_queries_per_block=64,
        vmem_limit_bytes=32 * 1024 * 1024,
        max_num_batched_tokens=512,
        max_num_seq=8,
        sliding_window: int | None = None,
        soft_cap: float | None = None,
    ):
        if not jtu.is_device_tpu_at_least(version=4):
            self.skipTest("Expect TPUv4+")
        cu_q_lens = [0]
        kv_lens = []
        for q_len, kv_len in seq_lens:
            assert q_len <= kv_len
            cu_q_lens.append(cu_q_lens[-1] + q_len)
            kv_lens.append(kv_len)

        max_num_batched_tokens = max(cu_q_lens[-1], max_num_batched_tokens)
        max_num_seq = max(len(seq_lens), max_num_seq)
        max_kv_len = max(kv_lens)
        pages_per_seq = ceil_div(max_kv_len, page_size)
        num_q_heads, num_kv_heads = num_heads

        cu_q_lens = jnp.array(cu_q_lens, dtype=jnp.int32)
        kv_lens = jnp.array(kv_lens, dtype=jnp.int32)
        cu_q_lens = jnp.pad(cu_q_lens,
                            (0, max_num_seq + 1 - cu_q_lens.shape[0]))
        kv_lens = jnp.pad(kv_lens, (0, max_num_seq - kv_lens.shape[0]))
        prng_key = jax.random.key(1234)
        k0, k1, k2 = jax.random.split(prng_key, 3)
        q = jax.random.normal(
            k0,
            (max_num_batched_tokens, num_q_heads, head_dim),
            dtype=dtype,
        )
        kv_pages = jax.random.normal(
            k1,
            (num_pages, page_size, num_kv_heads * 2, head_dim),
            dtype=dtype,
        )
        page_indices = jax.random.randint(k2, (max_num_seq, pages_per_seq),
                                          0,
                                          num_pages,
                                          dtype=jnp.int32)

        num_seqs = jnp.array([len(seq_lens)], dtype=jnp.int32)

        dynamic_validate_inputs(
            q,
            kv_pages,
            kv_lens,
            page_indices,
            cu_q_lens,
            num_seqs,
            sliding_window=sliding_window,
            soft_cap=soft_cap,
        )

        actual_num_q_tokens = cu_q_lens[num_seqs[0]]
        output = ragged_paged_attention(
            q,
            kv_pages,
            kv_lens,
            page_indices,
            cu_q_lens,
            num_seqs=num_seqs,
            num_kv_pages_per_block=min(num_kv_pages_per_block, pages_per_seq),
            num_queries_per_block=num_queries_per_block,
            vmem_limit_bytes=vmem_limit_bytes,
            sliding_window=sliding_window,
            soft_cap=soft_cap,
        )[:actual_num_q_tokens]

        expected = ref_ragged_paged_attention(
            q,
            kv_pages,
            kv_lens,
            page_indices,
            cu_q_lens,
            num_seqs=num_seqs,
            sliding_window=sliding_window,
            soft_cap=soft_cap,
        )
        tols = {
            "float32": 0.15,
            "bfloat16": 0.2,
        }
        tol = tols[jnp.dtype(dtype).name]
        self.assertAllClose(output, expected, atol=tol, rtol=tol)

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_basic(self, dtype):
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_decode_only(self, dtype):
        seq_lens = [
            (1, 18),
            (1, 129),
            (1, 597),
            (1, 122),
            (1, 64),
            (1, 322),
            (1, 463),
            (1, 181),
            (1, 1107),
            (1, 123),
            (1, 31),
            (1, 18),
            (1, 1229),
            (1, 229),
            (1, 87),
            (1, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_prefill_only(self, dtype):
        seq_lens = [
            (5, 18),
            (15, 129),
            (120, 597),
            (100, 122),
            (21, 64),
            (32, 322),
            (251, 463),
            (40, 181),
            (64, 1107),
            (99, 123),
            (10, 31),
            (5, 18),
            (3, 1229),
            (120, 229),
            (9, 87),
            (2, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
        )

    @parameterized.product(dtype=[jnp.float32, jnp.bfloat16], )
    def test_ragged_paged_attention_mixed(self, dtype):
        seq_lens = [
            (5, 18),
            (1, 129),
            (120, 597),
            (1, 122),
            (1, 64),
            (32, 322),
            (251, 463),
            (1, 181),
            (1, 1107),
            (99, 123),
            (1, 31),
            (5, 18),
            (3, 1229),
            (117, 229),
            (1, 87),
            (1, 1328),
        ]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
        )

    @parameterized.product(
        num_seqs=[1, 5, 16],
        # TODO(jevinjiang): Support more num_heads!
        num_heads=[(32, 8), (32, 16), (12, 2), (4, 4), (8, 1)],
        dtype=[jnp.float32, jnp.bfloat16],
        num_kv_pages_per_block=[4, 8],
        num_queries_per_block=[32, 64],
    )
    def test_ragged_paged_attention_complex(
        self,
        num_seqs,
        num_heads,
        dtype,
        num_kv_pages_per_block,
        num_queries_per_block,
    ):
        seq_lens = []
        for _ in range(num_seqs):
            q_len = random.randint(1, 100)
            kv_len = q_len + random.randint(0, 50)
            seq_lens.append((q_len, kv_len))
        # TODO(jevinjiang): Support non-128 head_dim!
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
            num_kv_pages_per_block=num_kv_pages_per_block,
            num_queries_per_block=num_queries_per_block,
        )

    @parameterized.product(
        num_kv_pages_per_block=[4, 8],
        num_queries_per_block=[32, 64],
        sliding_window=[None, 5, 128],
    )
    def test_ragged_paged_attention_sliding_window(
        self,
        num_kv_pages_per_block,
        num_queries_per_block,
        sliding_window: int | None,
    ):
        num_seqs = 5
        num_heads = (4, 4)
        dtype = jnp.float32
        seq_lens = []
        for _ in range(num_seqs):
            q_len = random.randint(1, 100)
            kv_len = q_len + random.randint(0, 50)
            seq_lens.append((q_len, kv_len))
        # TODO(jevinjiang): Support non-128 head_dim!
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
            num_kv_pages_per_block=num_kv_pages_per_block,
            num_queries_per_block=num_queries_per_block,
            sliding_window=sliding_window,
        )

    @parameterized.product(
        num_kv_pages_per_block=[4, 8],
        num_queries_per_block=[32, 64],
        soft_cap=[None, 50.0],
    )
    def test_ragged_paged_attention_logit_soft_capping(
        self,
        num_kv_pages_per_block,
        num_queries_per_block,
        soft_cap: float | None,
    ):
        num_heads = (12, 2)
        num_seqs = 2
        dtype = jnp.float32
        seq_lens = []
        for _ in range(num_seqs):
            q_len = random.randint(1, 100)
            kv_len = q_len + random.randint(0, 50)
            seq_lens.append((q_len, kv_len))
        head_dim = 128
        page_size = 16
        num_pages = 1000

        self._test_ragged_paged_attention(
            seq_lens,
            num_heads,
            head_dim,
            page_size,
            dtype,
            num_pages,
            num_kv_pages_per_block=num_kv_pages_per_block,
            num_queries_per_block=num_queries_per_block,
            soft_cap=soft_cap,
        )

    def test_ragged_paged_attention_sliding_window_should_be_positive(self):
        dtype = jnp.float32
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        with self.assertRaisesRegex(ValueError, "must be positive"):
            self._test_ragged_paged_attention(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                num_pages,
                sliding_window=0,
            )

        with self.assertRaisesRegex(ValueError, "must be positive"):
            self._test_ragged_paged_attention(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                num_pages,
                sliding_window=-1,
            )

    def test_ragged_paged_attention_soft_cap_cannot_be_zero(self):
        dtype = jnp.float32
        seq_lens = [(192, 328), (128, 180), (64, 255)]
        num_heads = (32, 8)
        head_dim = 128
        page_size = 16
        num_pages = 1000

        with self.assertRaisesRegex(ValueError, "must not be 0.0"):
            self._test_ragged_paged_attention(
                seq_lens,
                num_heads,
                head_dim,
                page_size,
                dtype,
                num_pages,
                soft_cap=0.0,
            )


if __name__ == "__main__":
    absltest.main(testLoader=jtu.JaxTestLoader())



#!/usr/bin/env python
# SPDX-License-Identifier: Apache-2.0
import os
import pathlib #
from typing import List
from setuptools import find_packages, setup

#ROOT_DIR = os.path.dirname(__file__)
ROOT_DIR = pathlib.Path(__file__).parent.resolve()


def get_path(*filepath) -> str:
    return os.path.join(ROOT_DIR, *filepath)


def get_requirements() -> List[str]:
    """Get Python package dependencies from requirements.txt."""
    requirements_path = ROOT_DIR / "requirements.txt"
    print(f"requirements_path: {requirements_path}")
    def _read_requirements(filename: str) -> List[str]:
        #with open(get_path(filename)) as f:
        with open(requirements_path) as f:
            requirements = f.read().strip().split("\n")
        resolved_requirements = []
        for line in requirements:
            if line.startswith("-r "):
                resolved_requirements += _read_requirements(line.split()[1])
            elif line.startswith("--"):
                continue
            else:
                resolved_requirements.append(line)
        return resolved_requirements

    try:
        requirements = _read_requirements("requirements.txt")
    except ValueError:
        print("Failed to read requirements.txt in vllm_tpu.")
    return requirements


print("Start the set up"),
print(f"Current directory: {pathlib.Path('.').resolve()}")
print(f"ROOT_DIR: {ROOT_DIR}")

setup(
    name="test-ylang",
    version=os.environ.get('VERSION'),
    description="",
    long_description=open("README.md").read() if hasattr(
        open("README.md"), "read") else "",
    long_description_content_type="text/markdown",
    author="tpu_commons Contributors",
    packages=find_packages(),
    python_requires=">=3.10",
    install_requires=get_requirements(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
print("Setup script finished.")



# SPDX-License-Identifier: Apache-2.0

#
# WARNING: This is a temporary compatibility module.
#
#
# THE PROBLEM:
# The ideal dependency injection pattern dictates that the "producer" of data
# (in this case, the vLLM engine) should be responsible for adapting its data
# into the abstract format that the "consumer" (the TPU worker) expects.
#
# However, this would require a simultaneous code change in both the `vllm` and
# `tpu_commons` repositories. Such cross-repository changes are difficult to
# coordinate, slow to land, and can easily cause breakages if the releases
# are not perfectly synchronized.
#
#
# THE TEMPORARY SOLUTION:
# To enable independent development and deployment, we are temporarily violating
# this pattern. We are making the consumer (`tpu_commons`) responsible for
# detecting and adapting the producer's raw data.
#
# This function checks if it has received a raw `vllm.SchedulerOutput` and,
# if so, wraps it in the appropriate adapter. This allows `vllm` to continue
# sending its raw data type without modification, decoupling the release cycles.
#
#
# THE FUTURE (HOW TO REMOVE THIS):
# This entire file should be deleted once the `vllm` repository has been updated.
# The required change in `vllm` is small and looks like this:
#
# --- SKELETON CODE FOR FUTURE vLLM CHANGE ---
# In the vLLM engine, where `execute_model` is called:
#
# from tpu_commons.adapters.vllm_adapters import VllmSchedulerOutputAdapter
# from vllm.v1.core.sched.output import SchedulerOutput
#
# # ... inside some method ...
#
# # OLD CODE:
# # concrete_work = SchedulerOutput(...)
# # self.tpu_backend.execute_model(concrete_work)
#
# # NEW CODE:
# concrete_work = SchedulerOutput(...)
# adapted_work = VllmSchedulerOutputAdapter(concrete_work)  # This line is added
# self.tpu_backend.execute_model(adapted_work)          # Pass the adapter
#
# --- END SKELETON CODE ---
#

import logging
from typing import Union

from vllm.lora.request import LoRARequest as VllmLoRARequest
# Import the concrete vLLM type for the check
from vllm.v1.core.sched.output import SchedulerOutput as VllmSchedulerOutput
from vllm.v1.kv_cache_interface import KVCacheConfig as VllmKVCacheConfig

from tpu_commons.adapters.vllm_adapters import (VllmKVCacheConfigAdapter,
                                                VllmLoRARequestAdapter,
                                                VllmSchedulerOutputAdapter)
from tpu_commons.di.abstracts import (AbstractKVCacheConfig,
                                      AbstractLoRARequest,
                                      AbstractSchedulerOutput)

logger = logging.getLogger(__name__)


def adapt_scheduler_output_if_needed(
    scheduler_output: Union[AbstractSchedulerOutput, VllmSchedulerOutput]
) -> AbstractSchedulerOutput:
    """
    Checks if the input is a raw VllmSchedulerOutput and wraps it.
    If it's already an AbstractSchedulerOutput, it's passed through.
    """
    if isinstance(scheduler_output, VllmSchedulerOutput):
        # logger.warning(
        #     "Received raw VllmSchedulerOutput. Performing temporary, on-the-fly "
        #     "adaptation. This is a compatibility feature and should be removed "
        #     "once the vLLM engine is updated to provide an adapted object.")
        return VllmSchedulerOutputAdapter(scheduler_output)

    if isinstance(scheduler_output, AbstractSchedulerOutput):
        return scheduler_output

    raise TypeError(
        f"Unsupported type for scheduler_output: {type(scheduler_output)}")


def adapt_kv_cache_config_if_needed(
    kv_cache_config: Union[AbstractKVCacheConfig, VllmKVCacheConfig]
) -> AbstractKVCacheConfig:
    """
    Checks if the input is a raw VllmKVCacheConfig and wraps it.
    If it's already an AbstractKVCacheConfig, it's passed through.
    """
    if isinstance(kv_cache_config, VllmKVCacheConfig):
        # logger.warning(
        #     "Received raw VllmKVCacheConfig. Performing temporary, on-the-fly "
        #     "adaptation. This is a compatibility feature and should be removed "
        #     "once the vLLM engine is updated to provide an adapted object.")
        return VllmKVCacheConfigAdapter(kv_cache_config)

    if isinstance(kv_cache_config, AbstractKVCacheConfig):
        return kv_cache_config

    raise TypeError(
        f"Unsupported type for kv_cache_config: {type(kv_cache_config)}")


def adapt_lora_request_if_needed(
    lora_request: Union[AbstractLoRARequest, VllmLoRARequest]
) -> AbstractLoRARequest:
    """
    Checks if the input is a raw VllmLoRARequest and wraps it.
    If it's already an AbstractLoRARequest, it's passed through.
    """
    if isinstance(lora_request, VllmLoRARequest):
        # logger.warning(
        #     "Received raw VllmLoRARequest. Performing temporary, on-the-fly "
        #     "adaptation. This is a compatibility feature and should be removed "
        #     "once the vLLM engine is updated to provide an adapted object.")
        return VllmLoRARequestAdapter(lora_request)

    if isinstance(lora_request, AbstractLoRARequest):
        return lora_request

    raise TypeError(f"Unsupported type for lora_request: {type(lora_request)}")



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Optional

from tpu_commons.adapters.vllm_adapters import (VllmLoRARequestAdapter,
                                                VllmSchedulerOutputAdapter)
from tpu_commons.di.interfaces import BackendInterface, HostInterface
from tpu_commons.worker.base import AbstractTpuWorker
from tpu_commons.worker.tpu_worker_jax import TPUWorker


class TPUBackend(BackendInterface):
    """
  The main entry point for the host system to interact with the TPU backend.

  This class implements the BackendInterface. It is responsible for creating
  and managing the concrete TPU worker instance and delegating calls to it.
  """

    def __init__(self,
                 host_interface: Optional[HostInterface] = None,
                 **worker_kwargs):
        """
        Initializes the TPUBackend.

        Args:
            host_interface: An optional object that implements the HostInterface,
                            providing a way for the backend to communicate with the host.
            **worker_kwargs: Additional keyword arguments to be passed to the
                            worker's constructor.
        """
        self.worker: AbstractTpuWorker = TPUWorker(
            host_interface=host_interface, **worker_kwargs)

    def launch_tpu_batch(self, batch_to_launch):
        """
        Launches a batch of requests on the TPU worker and returns the result.

        Args:
            batch_to_launch: The batch of requests to be processed.

        Returns:
            The result of the model execution.
        """
        adapted_batch = VllmSchedulerOutputAdapter(batch_to_launch)
        return self.worker.execute_model(adapted_batch)

    def add_lora(self, lora_request):
        """
        Adds a LoRA adapter to the worker.

        Args:
            lora_request: The LoRA request to be processed.
        """
        adapted_lora_request = VllmLoRARequestAdapter(lora_request)
        return self.worker.add_lora(adapted_lora_request)



# SPDX-License-Identifier: Apache-2.0
import os
from collections import defaultdict
from collections.abc import Sequence
from typing import Any, List, Tuple

import jax
import numpy as np
from jax._src import dtypes
from jax._src import mesh as mesh_lib
from jax._src import xla_bridge as xb
from jax._src.lib import xla_client as xc
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from vllm import envs

from tpu_commons.logger import init_logger

GBYTES = 1024 * 1024 * 1024
TPU_HEAD_SIZE_ALIGNMENT = 128
TPU_SECOND_LAST_MINOR = 8

_megacore = False
logger = init_logger(__name__)


def enable_megacore() -> None:
    global _megacore
    _megacore = True


def get_megacore() -> bool:
    return _megacore


def get_num_kv_heads_by_tp(num_kv_heads: int, tp_size: int) -> int:
    if tp_size <= num_kv_heads:
        assert num_kv_heads % tp_size == 0
        return num_kv_heads
    else:
        assert tp_size % num_kv_heads == 0
        return tp_size


def hbm_usage_bytes(devices: Any) -> List[Tuple[int, int]]:
    usage = []
    if envs.VLLM_TPU_USING_PATHWAYS:
        return pathways_hbm_usage_gb(devices)

    multihost_backend = os.environ.get("TPU_MULTIHOST_BACKEND", "").lower()
    if multihost_backend == "ray":
        # MemoryStats is only supported for addressable PjRt devices.
        # Assume all the devices have similar memory usage for now.
        # TODO(ranlihao): find a proper way to get the memory usage of each device.
        for device in devices:
            try:
                hbm_used = device.memory_stats()["bytes_in_use"]
                hbm_limit = device.memory_stats()["bytes_limit"]
                logger.info(
                    "Get memory stats for device %s. Assuming all devices have the same usage.",
                    device)
                usage.extend([(hbm_used, hbm_limit)] * len(devices))
                break
            except Exception as e:
                logger.warning(
                    "Failed to get memory stats for device %s: %s. ", device,
                    e)
    else:
        for device in devices:
            hbm_used = device.memory_stats()["bytes_in_use"]
            hbm_limit = device.memory_stats()["bytes_limit"]
            usage.append((hbm_used, hbm_limit))

    return usage


def pathways_hbm_usage_gb(devices: Any) -> List[Tuple[float, float]]:
    live_arrays = jax.live_arrays()
    hbm_used = defaultdict(int)
    # TODO(wenxindong): Find a way to get the accurate hbm limit on Pathways.
    hbm_limit = 33550237184
    for array in live_arrays:
        assert hasattr(array, 'sharding') and hasattr(
            array.sharding, 'device_set'
        ), "This function must not be called within jax tracer (e.g. jit, vmap, grad)"
        for device in array.sharding.device_set:
            hbm_used[device] += array.dtype.itemsize * array.size // len(
                array.sharding.device_set)
    return [(hbm_used[device], hbm_limit) for device in devices]


def hbm_usage_gb(devices: Any) -> List[Tuple[float, float]]:
    usage = hbm_usage_bytes(devices)
    usage = [(round(used / GBYTES, 2), round(limit / GBYTES, 2))
             for used, limit in usage]
    return usage


def get_padded_head_dim(head_dim: int) -> int:
    """Pads head_dim up to the nearest multiple of 128 for kernel performance."""
    return (head_dim + 127) // 128 * 128


def get_padded_num_heads(num_heads: int, sharding_size: int) -> int:
    if num_heads >= sharding_size:
        assert num_heads % sharding_size == 0
    else:
        assert sharding_size % num_heads == 0
        num_heads = sharding_size
    return num_heads


def get_dtype_packing(dtype):
    bits = dtypes.bit_width(dtype)
    return 32 // bits


def make_optimized_mesh(axis_shapes: Sequence[int],
                        axis_names: Sequence[str],
                        *,
                        devices: Sequence[xc.Device] | None = None):
    if devices is None:
        devices = xb.devices()

    def _is_1D(axis_shapes):
        return sum(x > 1 for x in axis_shapes) == 1

    if _is_1D(axis_shapes):
        dev_kind = devices[0].device_kind
        device_num = len(devices)
        if dev_kind == "TPU v6 lite":
            ordered_devices = None
            # NOTE(chengjiyao):
            # The coords of v6e-8 are
            # (0,0,0)
            # (1,0,0)
            # (0,1,0)
            # (1,1,0)
            # (0,2,0)
            # (1,2,0)
            # (0,3,0)
            # (1,3,0)
            if device_num == 8:
                ordered_devices = np.array([
                    devices[0],
                    devices[2],
                    devices[4],
                    devices[6],
                    devices[7],
                    devices[5],
                    devices[3],
                    devices[1],
                ])
            # NOTE(chengjiyao):
            # The coords of v6e-4 are
            # (0,0,0)
            # (1,0,0)
            # (0,1,0)
            # (1,1,0)
            elif device_num == 4:
                ordered_devices = np.array([
                    devices[0],
                    devices[2],
                    devices[3],
                    devices[1],
                ])
            if ordered_devices is not None:
                ordered_devices = np.array(ordered_devices)
                ordered_devices = ordered_devices.reshape(axis_shapes)
                mesh = mesh_lib.Mesh(ordered_devices, axis_names)
                logger.info("Use customized mesh: %s", mesh)
                return mesh

    return jax.make_mesh(axis_shapes, axis_names, devices=devices)


def device_array(mesh: Mesh, *args, sharding=None, **kwargs) -> jax.Array:
    """
    Create a device array with the specified mesh and sharding.

    Args:
        mesh: The JAX mesh to use for device placement
        *args: Positional arguments to pass to jax.device_put
        sharding: Optional sharding specification. If None, uses PartitionSpec(None)
        **kwargs: Keyword arguments to pass to jax.device_put

    Returns:
        A JAX array placed on the specified devices
    """
    if sharding is None:
        sharding = NamedSharding(mesh, PartitionSpec(None))
    return jax.device_put(*args, device=sharding, **kwargs)



import functools
import os
import random
from contextlib import nullcontext
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import jax
import jax.numpy as jnp
import jaxtyping
import numpy as np
import torch
import vllm.envs as envs
from flax import nnx
from torchax.ops.mappings import j2t_dtype
from vllm.config import VllmConfig
from vllm.distributed.kv_transfer import (get_kv_transfer_group,
                                          has_kv_transfer_group)
from vllm.forward_context import set_forward_context
from vllm.sequence import IntermediateTensors
from vllm.tasks import SupportedTask
from vllm.utils import cdiv
from vllm.v1.core.sched.output import SchedulerOutput as VllmSchedulerOutput
from vllm.v1.kv_cache_interface import KVCacheConfig
from vllm.v1.outputs import (EMPTY_MODEL_RUNNER_OUTPUT, DraftTokenIds,
                             ModelRunnerOutput)
from vllm.v1.request import Request
from vllm.v1.spec_decode.ngram_proposer import NgramProposer
from vllm.v1.worker.kv_connector_model_runner_mixin import \
    KVConnectorModelRunnerMixin

from tpu_commons import utils as common_utils
from tpu_commons.logger import init_logger
from tpu_commons.models.jax.attention_metadata import AttentionMetadata
from tpu_commons.models.jax.common.sharding import build_mesh
from tpu_commons.models.jax.layers.sample.rejection_sampler import \
    RejectionSampler
from tpu_commons.models.jax.layers.sample.sampling import (compute_logprobs,
                                                           gather_logprobs,
                                                           sample)
from tpu_commons.models.jax.layers.sample.sampling_metadata import \
    TPUSupportedSamplingMetadata
from tpu_commons.models.jax.model_loader import get_model
from tpu_commons.models.jax.utils.weight_utils import (
    shard_put, transfer_state_with_mappings)
from tpu_commons.runner import utils as runner_utils
from tpu_commons.runner.compilation_manager import CompilationManager
from tpu_commons.runner.input_batch_jax import CachedRequestState, InputBatch
from tpu_commons.runner.kv_cache_manager import KVCacheManager
from tpu_commons.runner.multimodal_manager import MultiModalManager
from tpu_commons.runner.persistent_batch_manager import PersistentBatchManager
from tpu_commons.runner.speculative_decoding_manager import \
    SpeculativeDecodingManager
from tpu_commons.runner.structured_decoding_manager import \
    StructuredDecodingManager
from tpu_commons.utils import device_array, make_optimized_mesh

logger = init_logger(__name__)

INVALID_TOKEN_ID = -1
# Smallest output size
MIN_NUM_SEQS = 8

DUMMY_METADATA = AttentionMetadata(
    input_positions=[],
    block_tables=[],
    request_distribution=[0, 0, 0],
)

TPU_STR_DTYPE_TO_TORCH_DTYPE = {
    "half": torch.half,
    "bfloat16": torch.bfloat16,
    "float": torch.float,
    "fp8": torch.float8_e4m3fn,
    "fp8_e4m3": torch.float8_e4m3fn,
    "fp8_e5m2": torch.float8_e5m2,
    "int8": torch.int8,
    "uint8": torch.uint8,
}


class TPUModelRunner(KVConnectorModelRunnerMixin):

    def __init__(
        self,
        vllm_config: VllmConfig,
        devices: List[Any],
    ):
        self.vllm_config = vllm_config
        self.model_config = vllm_config.model_config
        # TODO(jevinjiang): override block size based on RPA v3.
        self.cache_config = vllm_config.cache_config
        self.lora_config = vllm_config.lora_config
        self.load_config = vllm_config.load_config
        self.parallel_config = vllm_config.parallel_config
        self.scheduler_config = vllm_config.scheduler_config
        self.speculative_config = vllm_config.speculative_config
        self.observability_config = vllm_config.observability_config
        self.device_config = vllm_config.device_config

        self.devices = devices
        self.dtype = self.model_config.dtype
        self.maybe_forbid_compile = runner_utils.ForbidCompile(
        ) if envs.VLLM_XLA_CHECK_RECOMPILATION else nullcontext()

        self._init_random()
        self._init_mesh()
        self._init_phased_profiling()
        self._init_mm()
        self._init_speculative_decoding()
        self._init_inputs()

        # Delegate functions to specific manager classes.
        self.compilation_manager = CompilationManager(self)
        self.speculative_decoding_manager = SpeculativeDecodingManager(self)
        self.structured_decoding_manager = StructuredDecodingManager(self)
        self.kv_cache_manager = KVCacheManager(self)
        self.mm_manager = MultiModalManager(self)
        self.persistent_batch_manager = PersistentBatchManager(
            self.requests, self.input_batch, self.encoder_cache,
            self.uses_mrope, self.model_config)

        cache_config = self.cache_config
        if cache_config.cache_dtype == "auto":
            model_dtype = self.dtype
            if isinstance(model_dtype, str):
                self.kv_cache_dtype = TPU_STR_DTYPE_TO_TORCH_DTYPE[model_dtype]
            elif isinstance(getattr(model_dtype, 'dtype', None), jnp.dtype):
                self.kv_cache_dtype = j2t_dtype(model_dtype.dtype)
            elif isinstance(model_dtype, torch.dtype):
                self.kv_cache_dtype = model_dtype
            else:
                raise ValueError(
                    "KV cache is unsupported for model_dtype of %s",
                    model_dtype)
        else:
            self.kv_cache_dtype = TPU_STR_DTYPE_TO_TORCH_DTYPE[
                cache_config.cache_dtype]

    def _init_random(self):
        if self.model_config.seed is None:
            self.model_config.seed = 0
        random.seed(self.model_config.seed)
        np.random.seed(self.model_config.seed)
        self.rng_key = jax.random.key(self.model_config.seed)

    def _init_mesh(self) -> None:
        try:
            # TODO: Update override steps.
            sharding_strategy = \
                self.vllm_config.additional_config["sharding"]["sharding_strategy"]
        except KeyError:
            sharding_strategy = {"tensor_parallelism": len(self.devices)}

        if os.getenv("NEW_MODEL_DESIGN", False):
            self.mesh = build_mesh(self.devices, sharding_strategy)
        else:
            try:
                dp = sharding_strategy["data_parallelism"]
            except KeyError:
                dp = 1
            try:
                tp = sharding_strategy["tensor_parallelism"]
            except KeyError:
                tp = len(self.devices)

            axis_names = ("data", "model")
            mesh_shape = (dp, tp)

            self.mesh = make_optimized_mesh(mesh_shape,
                                            axis_names,
                                            devices=self.devices)
        logger.info(f"Init mesh | mesh={self.mesh}")

    def _init_phased_profiling(self) -> None:
        self.phased_profiling_dir = os.getenv("PHASED_PROFILING_DIR", "")
        self.phase_based_profiler = None
        if self.phased_profiling_dir:
            self.phase_based_profiler = runner_utils.PhasedBasedProfiler(
                self.phased_profiling_dir)

    def _init_mm(self) -> None:
        self.is_multimodal_model = None
        self.uses_mrope = self.model_config.uses_mrope

    def _init_speculative_decoding(self) -> None:
        if self.speculative_config:
            if self.speculative_config.method == "ngram":
                self.drafter = NgramProposer(self.vllm_config)
            else:
                raise NotImplementedError(
                    "Unsupported speculative decoding method: "
                    f"{self.speculative_config.method}")
            self.rejection_sampler = RejectionSampler()

    def _init_inputs(self) -> None:
        model_config = self.model_config
        cache_config = self.cache_config
        scheduler_config = self.scheduler_config

        self.sliding_window = model_config.get_sliding_window()
        self.block_size = cache_config.block_size
        self.max_model_len = model_config.max_model_len
        self.max_num_blocks_per_req = cdiv(self.max_model_len, self.block_size)
        # InputBatch needs to work with sampling tensors greater than padding
        # to avoid dynamic shapes. Also, avoid suboptimal alignment.
        self.max_num_reqs = max(scheduler_config.max_num_seqs, MIN_NUM_SEQS)
        # [16, 32, 64, 128, 256, 512, 1024, 2048]
        self.num_tokens_paddings = runner_utils.get_token_paddings(
            min_token_size=16,
            max_token_size=scheduler_config.max_num_batched_tokens,
            padding_gap=envs.VLLM_TPU_BUCKET_PADDING_GAP)
        # In case `max_num_tokens < max(num_tokens_paddings)` use the actual
        # padded max value to pre-allocate data structures and pre-compile.
        self.max_num_tokens = self.num_tokens_paddings[-1]

        # Request states.
        self.requests: dict[str, CachedRequestState] = {}
        # mm_hash ->  encoder_output
        self.encoder_cache: dict[str, jax.Array] = {}
        self.input_batch = InputBatch(
            max_num_reqs=self.max_num_reqs,
            max_model_len=self.max_model_len,
            max_num_batched_tokens=self.max_num_tokens,
            pin_memory=False,
            vocab_size=self.model_config.get_vocab_size(),
            block_sizes=[self.block_size],
            is_spec_decode=bool(self.vllm_config.speculative_config),
        )

        self.input_ids_cpu = np.zeros(self.max_num_tokens, dtype=np.int32)
        self.positions_cpu = np.zeros(self.max_num_tokens, dtype=np.int32)
        self.block_table_cpu = np.zeros(
            (self.max_num_reqs, self.max_num_blocks_per_req), dtype=np.int32)
        self.query_start_loc_cpu = np.zeros(self.max_num_tokens + 1,
                                            dtype=np.int32)
        self.seq_lens_cpu = np.zeros(self.max_num_tokens, dtype=np.int32)
        # Range tensor with values [0 .. self.max_num_tokens - 1].
        # Used to initialize positions / context_lens / seq_lens
        # Keep in int64 to avoid overflow with long context
        self.arange_cpu = np.arange(self.max_num_tokens, dtype=np.int64)
        self.num_reqs_paddings = runner_utils.get_req_paddings(
            min_req_size=MIN_NUM_SEQS, max_req_size=self.max_num_reqs)

        # Padding for logits. Without speculative decoding, each request has one position to select from.
        # With speculative decoding, each request has multiple positions to select from.
        max_logits_per_req = 1
        if self.speculative_config:
            max_logits_per_req = self.speculative_config.num_speculative_tokens + 1  # Including bonus token
            self.num_logits_paddings = runner_utils.get_token_paddings(
                min_token_size=MIN_NUM_SEQS,
                max_token_size=self.max_num_reqs * max_logits_per_req,
                padding_gap=0)
        else:
            self.num_logits_paddings = None

        self.temperatures_cpu = np.zeros(self.max_num_tokens, dtype=np.float32)
        self.top_ps_cpu = np.zeros(self.max_num_tokens, dtype=np.float32)
        self.top_ks_cpu = np.zeros(self.max_num_tokens, dtype=np.int32)

        # tensors for structured decoding
        self.vocab_size = self.model_config.get_vocab_size()
        self.grammar_bitmask_cpu = np.zeros(
            (self.max_num_reqs, cdiv(self.vocab_size, 32)),
            dtype=np.int32,
        )
        self.require_structured_out_cpu = np.zeros(
            (self.max_num_reqs, 1),
            dtype=np.bool_,
        )
        self.structured_decode_arange = np.arange(0, 32, dtype=np.int32)

        # multi-modal support
        # Only relevant for models using M-RoPE (e.g, Qwen2-VL)

        # NOTE: When M-RoPE is enabled, position ids are 3D regardless of
        # the modality of inputs. For text-only inputs, each dimension has
        # identical position IDs, making M-RoPE functionally equivalent to
        # 1D-RoPE.
        # See page 5 of https://arxiv.org/abs/2409.12191
        self.mrope_positions_cpu = np.zeros((3, self.max_num_tokens),
                                            dtype=np.int64)

    def load_model(self):
        self.model_fn, self.compute_logits_fn, self.get_multimodal_embeddings_fn, self.get_input_embeddings_fn, self.state = get_model(
            self.vllm_config,
            self.rng_key,
            self.mesh,
        )
        self.rng_params_for_sampling = nnx.Rngs(
            jax.random.key(self.model_config.seed)).params()
        self.is_multimodal_model = (self.model_config.is_multimodal_model
                                    and self.get_multimodal_embeddings_fn
                                    is not None)

        logger.info(f"Init model | "
                    f"hbm={common_utils.hbm_usage_gb(self.devices)}Gb")

    def get_supported_tasks(self) -> tuple[SupportedTask, ...]:
        return ("generate", )

    def get_kv_cache_spec(self):
        return self.kv_cache_manager.get_kv_cache_spec()

    def initialize_kv_cache(self, kv_cache_config: KVCacheConfig) -> None:
        self.kv_caches = []
        self.kv_cache_manager.initialize_kv_cache(kv_cache_config)
        if has_kv_transfer_group():
            get_kv_transfer_group().register_runner(self)

    def capture_model(self) -> None:
        self.compilation_manager.capture_model()

    def execute_model(
        self,
        scheduler_output: "VllmSchedulerOutput",
        intermediate_tensors: Optional[IntermediateTensors] = None,
    ) -> ModelRunnerOutput:
        return self._execute_model(scheduler_output)[1]

    def _execute_model(
        self,
        scheduler_output: "VllmSchedulerOutput",
    ) -> tuple[AttentionMetadata, ModelRunnerOutput]:
        self.persistent_batch_manager.update_states(scheduler_output)
        if not scheduler_output.total_num_scheduled_tokens:
            if has_kv_transfer_group():
                return DUMMY_METADATA, self.kv_connector_no_forward(
                    scheduler_output, self.vllm_config)

            # Return empty ModelRunnerOutput if there's no work to do.
            # TODO(fhzhang): We rely on empty cycles to remove requests in input batch. Fix it to reduce overhead.
            logger.debug(f"Nothing scheduled: {scheduler_output}!")
            # NOTE(pooyam): There is no guarantee that scheduler is not sending empty output: https://github.com/vllm-project/vllm/blob/7cfea0df390c154c1026f77d3682e2733ca4aca8/vllm/v1/engine/core.py#L275
            # Why they are not preventing that is not clear to me.
            if len(scheduler_output.finished_req_ids) == 0:
                logger.warning(
                    "Should not schedule a request that does nothing!")
                # raise Exception(
                #     "Should not schedule a request that does nothing!")
            return DUMMY_METADATA, EMPTY_MODEL_RUNNER_OUTPUT,

        (input_ids, attn_metadata, sampling_metadata, logits_indices,
         spec_decode_metadata) = self._prepare_inputs(scheduler_output)

        # multi-modal support
        if self.is_multimodal_model:
            # Run the multimodal encoder if any.
            # We have the modality embeds at this time.
            self.mm_manager.execute_mm_encoder(scheduler_output)
            mm_embeds = self.mm_manager.gather_mm_embeddings(scheduler_output)
        else:
            mm_embeds = []

        # NOTE(Wenlong): For multi-modal model,
        # it will embed the text tokens and merge with the existing modality embeds
        # Later, the multi-modality model will take the embedding as the input.
        # For text-only model, this does nothing. It will input the input_ids and
        # leave the mebedding job inside the forward pass
        input_ids, inputs_embeds = self._get_input_ids_embeds(
            input_ids, mm_embeds)

        # TODO: Disable this for now
        if self.is_multimodal_model:
            self.maybe_forbid_compile = nullcontext()

        # TODO: make _get_input_ids_embeds within this context
        # NOTE: right now, mm model will use embeddings as the input,
        # but text-only model will use input_ids
        with self.maybe_forbid_compile:

            with set_forward_context(
                    None,
                    self.vllm_config,
            ), self.maybe_get_kv_connector_output(
                    scheduler_output) as kv_connector_output:
                # NOTE(Wenlong): It takes both `input_ids` and `inputs_embeds`,
                # but one of them would be `None`
                self.kv_caches, hidden_states = self.model_fn(
                    self.state,
                    self.kv_caches,
                    input_ids,
                    attn_metadata,
                    inputs_embeds,
                    tuple(self.layer_name_to_kvcache_index.items()),
                )

            hidden_states = self._select_from_array_fn(hidden_states,
                                                       logits_indices)
            logits = self.compute_logits_fn(self.state, hidden_states)
            if scheduler_output.grammar_bitmask is not None:
                (
                    require_struct_decoding, grammar_bitmask_padded, arange
                ) = self.structured_decoding_manager.prepare_structured_decoding_input(
                    logits, scheduler_output)
                logits = self.structured_decoding_manager.structured_decode_fn(
                    require_struct_decoding,
                    grammar_bitmask_padded,
                    logits,
                    arange,
                )
            tpu_sampling_metadata = sampling_metadata
            if spec_decode_metadata is None:
                next_tokens = sample(
                    self.rng_params_for_sampling,
                    self.mesh,
                    logits,
                    tpu_sampling_metadata,
                )
            else:
                bonus_logits = self._select_from_array_fn(
                    logits, spec_decode_metadata.bonus_logits_indices)
                bonus_token_ids = sample(
                    self.rng_params_for_sampling,
                    self.mesh,
                    bonus_logits,
                    tpu_sampling_metadata,
                )
                target_logits = self._select_from_array_fn(
                    logits, spec_decode_metadata.target_logits_indices)
                next_tokens = self.rejection_sampler(
                    draft_token_ids=spec_decode_metadata.draft_token_ids,
                    num_draft_tokens=spec_decode_metadata.draft_lengths,
                    draft_probs=None,
                    target_logits=target_logits,
                    bonus_token_ids=bonus_token_ids,
                    sampling_metadata=tpu_sampling_metadata,
                    key=self.rng_params_for_sampling,
                )

            if tpu_sampling_metadata.logprobs:
                logprobs = self._compute_and_gather_logprobs(
                    logits, next_tokens, self.model_config.max_logprobs)
            else:
                logprobs = None

        num_reqs = self.input_batch.num_reqs

        # Update the cache state concurrently. Code above will not block until
        # we use `selected_token_ids`. Add mark_step if post-processing changes
        request_seq_lens: list[tuple[int, CachedRequestState, int]] = []
        discard_sampled_tokens_req_indices = []
        for i, req_id in zip(range(num_reqs), self.input_batch.req_ids):
            assert req_id is not None
            req_state = self.requests[req_id]
            seq_len = (req_state.num_computed_tokens +
                       scheduler_output.num_scheduled_tokens[req_id])
            if seq_len >= req_state.num_tokens:
                request_seq_lens.append((i, req_state, seq_len))
            else:
                # Ignore the sampled token from the partial request.
                # Rewind the generator state as if the token was not sampled.
                generator = self.input_batch.generators.get(i)
                if generator is not None:
                    # This relies on cuda-specific torch-internal impl details
                    generator.set_offset(generator.get_offset() - 4)

                # Record the index of the request that should not be sampled,
                # so that we could clear the sampled tokens before returning.
                discard_sampled_tokens_req_indices.append(i)

        assert all(
            req_id is not None for req_id in
            self.input_batch.req_ids[:num_reqs]), "req_ids contains None"
        req_ids = cast(list[str], self.input_batch.req_ids[:num_reqs])

        prompt_logprobs_dict = {}
        for req_id in self.input_batch.req_ids[:num_reqs]:
            prompt_logprobs_dict[req_id] = None

        if spec_decode_metadata is None:
            next_tokens = np.asarray(jax.device_get(next_tokens))
            selected_token_ids = np.expand_dims(next_tokens[:num_reqs], 1)
            valid_sampled_token_ids = selected_token_ids.tolist()
        else:
            valid_sampled_token_ids = self.rejection_sampler.parse_output(
                next_tokens, self.input_batch.vocab_size,
                spec_decode_metadata.draft_lengths_cpu, num_reqs,
                spec_decode_metadata.draft_token_ids.shape[0])

        # Mask out the sampled tokens that should not be sampled.
        for i in discard_sampled_tokens_req_indices:
            valid_sampled_token_ids[i].clear()
        # Append sampled tokens
        for req_idx, req_state, _ in request_seq_lens:
            sampled_ids = valid_sampled_token_ids[req_idx]
            if not sampled_ids:
                continue

            start_idx = self.input_batch.num_tokens_no_spec[req_idx]
            end_idx = start_idx + len(sampled_ids)
            assert end_idx <= self.max_model_len, (
                "Sampled token IDs exceed the max model length. "
                f"Total number of tokens: {end_idx} > max_model_len: "
                f"{self.max_model_len}")

            self.input_batch.token_ids_cpu[req_idx,
                                           start_idx:end_idx] = sampled_ids
            self.input_batch.num_tokens_no_spec[req_idx] = end_idx
            self.input_batch.num_tokens[req_idx] = end_idx
            req_state.output_token_ids.extend(sampled_ids)

        if logprobs is not None:
            logprobs_lists = logprobs.tolists()
        else:
            logprobs_lists = None

        if self.speculative_config:
            self.speculative_decoding_manager.propose_draft_token_ids(
                valid_sampled_token_ids)

        model_runner_output = ModelRunnerOutput(
            req_ids=req_ids,
            req_id_to_index=self.input_batch.req_id_to_index,
            sampled_token_ids=valid_sampled_token_ids,
            logprobs=logprobs_lists,
            prompt_logprobs_dict=prompt_logprobs_dict,
            pooler_output=[],
            kv_connector_output=kv_connector_output,
        )
        return attn_metadata, model_runner_output

    @functools.partial(jax.jit, static_argnums=(0, ))
    def _select_from_array_fn(self, array, indices_to_select):
        return array[indices_to_select]

    @staticmethod
    @functools.partial(jax.jit, static_argnames=("max_logprobs", ))
    def _compute_and_gather_logprobs(logits, next_tokens, max_logprobs):
        logprobs = compute_logprobs(logits)
        return gather_logprobs(logprobs, next_tokens, max_logprobs)

    def _prepare_inputs(self, scheduler_output: "VllmSchedulerOutput"):
        total_num_scheduled_tokens = scheduler_output.total_num_scheduled_tokens
        assert total_num_scheduled_tokens > 0
        num_reqs = self.input_batch.num_reqs
        assert num_reqs > 0

        # Get the number of scheduled tokens for each request.
        num_scheduled_tokens_per_req = []
        max_num_scheduled_tokens_all_reqs = 0
        for req_id in self.input_batch.req_ids[:num_reqs]:
            assert req_id is not None
            num_tokens = scheduler_output.num_scheduled_tokens[req_id]
            num_scheduled_tokens_per_req.append(num_tokens)
            max_num_scheduled_tokens_all_reqs = max(
                max_num_scheduled_tokens_all_reqs, num_tokens)
        num_scheduled_tokens_per_req = np.array(num_scheduled_tokens_per_req,
                                                dtype=np.int32)
        assert max_num_scheduled_tokens_all_reqs > 0

        # Get request indices.
        # E.g., [2, 5, 3] -> [0, 0, 1, 1, 1, 1, 1, 2, 2, 2]
        # For each scheduled token, what are the corresponding req index.
        req_indices = np.repeat(self.arange_cpu[:num_reqs],
                                num_scheduled_tokens_per_req)

        # Get batched arange.
        # E.g., [2, 5, 3] -> [0, 1, 0, 1, 2, 3, 4, 0, 1, 2]
        # For each scheduled token, what is its position in corresponding req.
        arange = np.concatenate(
            [self.arange_cpu[:n] for n in num_scheduled_tokens_per_req])

        # Get positions.
        positions_np = self.positions_cpu[:total_num_scheduled_tokens]
        np.add(self.input_batch.num_computed_tokens_cpu[req_indices],
               arange,
               out=positions_np)

        # Multi-modal support
        # Calculate M-RoPE positions.
        # Only relevant for models using M-RoPE (e.g, Qwen2-VL)
        if self.uses_mrope:
            self.mm_manager.calc_mrope_positions(scheduler_output)

        # Get token indices.
        # E.g., [0, 1, 0, 1, 2, 3, 4, 0, 1, 2]
        # -> [0, 1, M, M + 1, M + 2, M + 3, M + 4, 2 * M, 2 * M + 1, 2 * M + 2]
        # where M is the max_model_len.
        token_indices = (positions_np +
                         req_indices * self.input_batch.token_ids_cpu.shape[1])

        # NOTE(woosuk): We use torch.index_select instead of np.take here
        # because torch.index_select is much faster than np.take for large
        # tensors.
        np.take(self.input_batch.token_ids_cpu.flatten(),
                token_indices,
                out=self.input_ids_cpu[:total_num_scheduled_tokens])

        # Prepare the attention metadata.
        self.query_start_loc_cpu[0] = 0
        np.cumsum(num_scheduled_tokens_per_req,
                  out=self.query_start_loc_cpu[1:num_reqs + 1])
        self.query_start_loc_cpu[num_reqs + 1:] = 1

        self.seq_lens_cpu[:num_reqs] = (
            self.input_batch.num_computed_tokens_cpu[:num_reqs] +
            num_scheduled_tokens_per_req)

        # Do the padding and copy the tensors to the TPU.
        padded_total_num_scheduled_tokens = runner_utils.get_padded_token_len(
            self.num_tokens_paddings, total_num_scheduled_tokens)
        # Zero out to avoid spurious values from prev iteration (last cp chunk)
        self.input_ids_cpu[
            total_num_scheduled_tokens:padded_total_num_scheduled_tokens] = 0

        # Please see runner_utils.PhasedBasedProfiler for details
        if self.phase_based_profiler:
            batch_composition_stats = runner_utils.get_batch_composition_stats(
                self.input_batch, total_num_scheduled_tokens, num_reqs,
                padded_total_num_scheduled_tokens, scheduler_output)

            self.phase_based_profiler.step(batch_composition_stats)

        # Inputs
        input_ids = self.input_ids_cpu[:padded_total_num_scheduled_tokens]
        positions = self.positions_cpu[:padded_total_num_scheduled_tokens]
        mrope_positions = self.mrope_positions_cpu[:, :
                                                   padded_total_num_scheduled_tokens]
        block_tables = self.block_table_cpu[:self.max_num_reqs]
        block_tables[:num_reqs, :self.max_num_blocks_per_req] = (
            self.input_batch.block_table[0].get_cpu_tensor()[:num_reqs])
        query_start_loc = self.query_start_loc_cpu[:self.max_num_reqs + 1]
        seq_lens = self.seq_lens_cpu[:self.max_num_reqs]
        request_distribution = np.array(self.input_batch.request_distribution)
        padded_num_reqs = runner_utils.get_padded_num_reqs_with_upper_limit(
            num_reqs, self.max_num_reqs)
        use_spec_decode = len(
            scheduler_output.scheduled_spec_decode_tokens) > 0
        if not use_spec_decode:
            logits_indices = self.query_start_loc_cpu[1:padded_num_reqs +
                                                      1] - 1
            spec_decode_metadata = None
        else:
            num_draft_tokens = np.zeros(num_reqs, dtype=np.int32)
            for req_id, draft_token_ids in (
                    scheduler_output.scheduled_spec_decode_tokens.items()):
                req_idx = self.input_batch.req_id_to_index[req_id]
                num_draft_tokens[req_idx] = len(draft_token_ids)

            spec_decode_metadata = self.speculative_decoding_manager.get_spec_decode_metadata(
                num_draft_tokens, self.query_start_loc_cpu[1:num_reqs + 1],
                padded_num_reqs)
            logits_indices = spec_decode_metadata.final_logits_indices

        # Put to device
        sampling_metadata = TPUSupportedSamplingMetadata.from_input_batch(
            self.mesh, self.input_batch, padded_num_reqs)
        if self.uses_mrope:
            positions = mrope_positions

        # Convert block_tables to 1D on cpu.
        block_tables = block_tables.reshape(-1)

        (input_ids, positions, block_tables, query_start_loc, seq_lens,
         logits_indices, request_distribution) = device_array(
             self.mesh, (input_ids, positions, block_tables, query_start_loc,
                         seq_lens, logits_indices, request_distribution))

        return (
            input_ids,
            AttentionMetadata(
                input_positions=positions,
                block_tables=block_tables,
                seq_lens=seq_lens,
                query_start_loc=query_start_loc,
                request_distribution=request_distribution,
            ),
            sampling_metadata,
            logits_indices,
            spec_decode_metadata,
        )

    def _get_input_ids_embeds(self, input_ids: jax.Array,
                              mm_embeds: list[jax.Array]):
        if self.is_multimodal_model:
            inputs_embeds = self.get_input_embeddings_fn(
                self.state,
                input_ids=input_ids,
                multimodal_embeddings=mm_embeds,
            )
            return None, inputs_embeds
        else:
            return input_ids, None

    def take_draft_token_ids(self) -> Optional[DraftTokenIds]:
        return self.speculative_decoding_manager.take_draft_token_ids()

    ###### Local disagg utilities ######

    def get_kv_cache_for_block_ids(
        self,
        block_ids: List[int],
    ) -> List[jax.Array]:
        return self.kv_cache_manager.get_kv_cache_for_block_ids(block_ids)

    def transfer_kv_cache(self,
                          kv_cache_slices: List[jax.Array]) -> List[jax.Array]:
        return self.kv_cache_manager.transfer_kv_cache(kv_cache_slices)

    def insert_request_with_kv_cache(
        self,
        request: "Request",
        kv_cache_slices: List[jax.Array],
        block_ids: List[List[int]],
    ):
        return self.kv_cache_manager.insert_request_with_kv_cache(
            request, kv_cache_slices, block_ids)

    ###### RL framework integration ######

    def _sync_weights(
        self,
        updated_weights: jaxtyping.PyTree,
        mappings: Dict[str, Tuple[str, Tuple[str]]],
        transpose_keys: Dict[str, Tuple[int]],
        reshard_fn: Callable[[jaxtyping.PyTree, jaxtyping.PyTree],
                             jaxtyping.PyTree] = None
    ) -> None:
        """For RL framework integration."""
        if reshard_fn is not None:
            updated_weights = reshard_fn(updated_weights, self.state)
            shard = None
        else:
            shard = functools.partial(shard_put, mesh=self.mesh)
        self.state = transfer_state_with_mappings(
            src_state=updated_weights,
            tgt_state=self.state,
            mappings=mappings,
            transpose_keys=transpose_keys,
            shard=shard)
